-- Action Bar Profiles
PAB = LibStub("AceAddon-3.0"):NewAddon("ProfilesActionBars", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "LibNotify-1.0")

PABversion = "v2.0-Beta"
ProfilesActionBarsForWoWversion = 70100
FORCEPAB = nil -- If 1, then will load even when versions mismatch
PABloadSpeed = nil

local AcquireFrame, ReleaseFrame
do
	local frame_cache = {}
	
	function AcquireFrame(parent)
		local frame = tremove(frame_cache) or CreateFrame("Frame")
		frame:SetParent(parent)
		return frame
	end
	
	function ReleaseFrame(frame)
		frame:Hide()
		frame:SetParent(nil)
		frame:ClearAllPoints()
		tinsert(frame_cache, frame)
	end
end    -- do block

function PAB:OnInitialize()
	PABdb = LibStub("AceDB-3.0"):New("ProfilesActionBars", nil, "Default")
	PABglobaldb = LibStub("AceDB-3.0"):New("ProfilesActionBarsG", nil, "Default")
	PABglobaldb.global.minimap = PABglobaldb.global.minimap or {} -- Thanks for spotting it Mik_lie@curse
	ABicon:Register("ProfilesActionBarsMENU", ABbunnyLDB, PABglobaldb.global.minimap)
	PAB:SetupOptions()
	PROFILES_TYST = PABdb.char.PROFILES_TYST
	
	
	if PABdb.char.PABLoadProfileAtStart == 1 then
		PAB:RegisterEvent("PLAYER_ENTERING_WORLD")
	end
	
	--- Configure LibNotify
	PAB:SetNotifyIcon("Interface\\Spellbook\\Spellbook-Icon")
	PABdb.profile["notifications"] = {}
	PAB:SetNotifyStorage(PABdb.profile["notifications"])
	PAB:Notify("Profiles: Actionbars v"..PABversion.." loaded", "")
	
	PABloadSpeed = PABglobaldb.global.PABloadSpeed or 0.3
	PABglobaldb.global.PABloadSpeed = PABloadSpeed

end

function PAB:VersCh() -- Main version block
	if PABdb then PABdb.char.PABLoadProfileAtStart = 0 end
	PAB:Echo("Version mismatch.Please update it from curse.com or run /pab forceload")
end

function PAB:ForceLoad()
	
	if FORCEPAB then
		FORCEPAB = nil
		PAB:Echo("Forcing Actiobars load")
		PAB:OnInitialize()
	else FORCEPAB = 1
	end
end

function PAB:PLAYER_ENTERING_WORLD(event, addonname) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	PAB:UnregisterEvent("PLAYER_ENTERING_WORLD")
	
	if PABdb.char.Active_profile and PABdb.char.Active_profile ~= "Restore [LOCKED]" then
		PAB:PreLoadABProfile(PABdb.char.Active_profile)
	end
	
	
	if (not PABdb.char.PABshow_minimap_icon) then
		ABicon:Show("Bunnies!")
	else
		ABicon:Hide("Bunnies!")
	end
end

-- ProfilesHub: Send message to all slaves that we are now exporting profile
function PAB:PreExport(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:SendAceMsg("export", name) -- ProfilesHub
end



-- This is an alias for ProfilesHub
function PAB:Export(...) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:ExportCharProfile(...)
end

-- Export char profile to global
function PAB:ExportCharProfile(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if name == nil then
		name = PABdb.char.Active_profile
	end
	
	if PAB:ExistsProfile(name) then
		local orig_profile = PABdb:GetCurrentProfile()
		
		PABglobaldb:SetProfile("General: " .. name)
		PABdb:SetProfile(name)
		
		-- Add class to db
		-- Lets us warn user if he tries to load a rogue profile on his druid
		local counter_addclass = 0
		
		-- MPglobaldb.profile = PABdb.profile
		for index, value in orderedPairs(PABdb.profile) do
			PABglobaldb.profile[index] = value
			counter_addclass = index + 1
		end
		
		-- Add class to db
		PABglobaldb.profile[counter_addclass + 1] = { select(2, UnitClass('player')), 0, 0 }
		
		PABdb:SetProfile(orig_profile)
		
		return 1
	else
		PAB:Echo("No such profile exists: " .. name)
		return nil
	end
end

function PAB:Echo(text)
	if PROFILES_TYST then return end
	DEFAULT_CHAT_FRAME:AddMessage("|cFFFF00FFProfiles: Action bars - " .. text)
end

-- ProfilesHub: Send message to all slaves that we are now saving profile
function PAB:PreSaveABProfile(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:SendAceMsg("save", name) -- ProfilesHub
end

-- This is an alias for ProfilesHub
function PAB:Save(...) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:SaveABSProfile(...)
end

-- This is an alias for ProfilesHub
function PAB:Load(...) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:LoadABSProfile(...)
end


-- ProfilesHub: Send message to all slaves that we are now switching profile
function PAB:PreLoadABProfile(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	
	-- No such profile saved
	if not PAB:ExistsProfile(name) then PAB:Echo(name .. " profile not found.") return end
	
	-- In combat!
	if InCombatLockdown() then PAB:Echo("Cannot change macros in combat!") return end
	
	PAB:SendAceMsg("load", name) -- ProfilesHub
end

--- Usage: spellBookIndexbyName("Fireball","PET") = 5
---
--- Usage: spellBookIndexbyName("Auto Attack") = 2
function PAB:spellBookIndexbyName(query, spellBook)
	local spellName, last
	
	spellBook = spellBook or "SPELL"
	
	if spellBook == "SPELL" then
		for index = 1, 500, 1 do
			spellName = GetSpellBookItemName(index, BOOKTYPE_SPELL)
			
			if not spellName then break end
			
			if spellName ~= last then
				last = spellName
				
				if spellName == query then return index end
			end
		end -- </BOOKTYPE_SPELL iteration>
	end
	
	
	if spellBook == "PET" then
		for index = 1, 500, 1 do
			spellName = GetSpellBookItemName(index, BOOKTYPE_PET)
			
			if not spellName then break end
			
			if spellName ~= last then
				last = spellName
				
				
				spellName = spellName:lower()
				spellName = spellName:gsub("PET_ACTION_", "")
				spellName = spellName:gsub("_", " ")
				
				if spellName == query then return index end
			end
		end -- </BOOKTYPE_PET iteration>
	end
end

--- Usage: PAB:GetMountIndex(5784) -> 17 -> C_MountJournal.Pickup(17) (Felsteed)
--- Used for OLD ID
--- @param mountSpellID '5784' Spell ID of mount
--- @return '17' MountIndex from for -> C_MountJournal.Pickup(i) -> C_MountJournal.GetMountInfoByID
local mountsCache = {}
local mountsCacheReversed = {} -- PAB:GetMountIndex(19) -> 5784 -> Iterate from this number

function PAB:GetMountIndex(mountSpellIDInput)
	-- Cache all the mounts
	-- So you can do GetActionInfo(280) -> C_MountJournal.Pickup(446) conversion
	-- to mount Traveler's Tundra Mammoth
	
	C_MountJournal.SetCollectedFilterSetting(1, true)
	C_MountJournal.SetCollectedFilterSetting(2, false)
	
	
	if not mountSpellIDInput then
		return nil
	elseif mountsCache[mountSpellIDInput] then
		return mountsCache[mountSpellIDInput]
	elseif mountSpellIDInput == 268435455 or mountSpellIDInput == 0 then
		return 0
	else
		local resumeAt = table.getn(mountsCacheReversed)
		for i = resumeAt, C_MountJournal.GetNumMounts() do
			ClearCursor()
			C_MountJournal.Pickup(i)
			local _, cnum = GetCursorInfo()
			ClearCursor()
			if pcall(function() C_MountJournal.GetMountInfoByID(cnum) end) then
				local mountSpellaltID = (select(2,C_MountJournal.GetMountInfoByID(cnum)))
				
				mountsCacheReversed[i] = "1"
				if mountSpellaltID and not mountsCache[mountSpellaltID] then
					mountsCache[mountSpellaltID] = i
				end
				
				if mountSpellIDInput == mountSpellaltID then
					return i
				end
			end
		end
	end
end



-- if PetHasActionBar() then -- hasPet
-- for petActionIndex =1,NUM_PET_ACTION_SLOTS do
-- PAB:SavePetSpell((select(1,GetPetActionInfo(petActionIndex))),petActionIndex)
-- end
-- end

--- Usage: PAB:pickupItem(1,"save",["placeAction"])
---
--- Usage: PAB:pickupItem(1,"load",["placeAction"])
---
--- @return Save() PABdb.profile[1] = { spellID, altID (for toys),icon, version, name, abPos,	}
--- @return Load() PickupItem(6603) (place action if testing)
--- @param actionbarPos	- Actionbars position
--- @param save/load	- Save() to profile[1] / Loads() from profile[1]
--- Description: For easier overview of save and load, converted them into 1 function.
--- 			 Also because the error handling is getting bigger and bigger
function PAB:pickupItem(abPos, command) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if command == "save" then
		local SavedItemID 		= (select(2,GetActionInfo(abPos)))
		local SavedItemType 	= (select(1, GetActionInfo(abPos)))
		local SavedItemName 	= (select(1, GetItemInfo(SavedItemID)))
		local SavedItemIcon 	= GetActionTexture(abPos)
		
		PABdb.profile[abPos] = {}
		PABdb.profile[abPos]["type"] 		= SavedItemType
		PABdb.profile[abPos]["ID"] 			= SavedItemID
		PABdb.profile[abPos]["abPos"] 		= abPos
		PABdb.profile[abPos]["name"] 		= SavedItemName
		PABdb.profile[abPos]["version"] 	= ProfilesActionBarsForWoWversion
		PABdb.profile[abPos]["icon"] 		= SavedItemIcon
	end -- </save>
	
	
	
	if command == "load" then
		-- Regular Load() ITEM
		if pcall(function() PickupItem(PABdb.profile[abPos]["ID"]) end) then
			ClearCursor()
			
			PickupItem( PABdb.profile[abPos]["ID"] )
		end
		
		-- Old load()
		if  not GetCursorInfo() and PABdb.profile[abPos][2] and pcall(function() PickupItem(PABdb.profile[abPos][2]) end) then
			ClearCursor()
			
			PickupItem( PABdb.profile[abPos][2] )
			PAB:NotifyOnce("'"..PABdb:GetCurrentProfile().."': Old profile!", "Warning: This profile contains old records that cannot be loaded in case of version change. |cffee4400Please re-save the profile|r.")
		end
		
		-- Name Load() TOYS
		if  not GetCursorInfo() and PABdb.profile[abPos]["name"] then
			ClearCursor()
			
			for ToyIndex = 1, C_ToyBox.GetNumToys() do
				local SavedToyID 	= C_ToyBox.GetToyFromIndex(ToyIndex)
				
				if SavedToyID and PlayerHasToy(   SavedToyID  ) then
					local SavedToyName = (select(2, C_ToyBox.GetToyInfo(SavedToyID)))
					
					if PABdb.profile[abPos]["name"] == SavedToyName and pcall(function() C_ToyBox.PickupToyBoxItem(SavedToyID) end) then
						ClearCursor()
						C_ToyBox.PickupToyBoxItem(SavedToyID)
					end
				end
			end
		end
		
		-- Name Load() ITEMS
		if  not GetCursorInfo() and PABdb.profile[abPos]["name"] then
			ClearCursor()
			
			-- bags
			for bagNumber = -1, 11 do
				local nSlots = GetContainerNumSlots(bagNumber)
				
				for slotNumber = 1, nSlots do
					local SavedItemID 	= GetContainerItemID(bagNumber, slotNumber)
					
					if SavedItemID then
						local SavedItemName = (select(1, GetItemInfo(SavedItemID)))
						
						if PABdb.profile[abPos]["name"] == SavedItemName and pcall(function() PickupItem(SavedItemID) end) then
							ClearCursor()
							
							PickupItem(SavedItemID)
						
						end
					end
				end
			end
			
			-- charslots
			if  not GetCursorInfo() then
				
				local nSlots = 19
				for charSlotNumber = 0, nSlots do
					local SavedItemID 	= GetInventoryItemID("player", charSlotNumber)
					
					if SavedItemID then
						local SavedItemName = (select(1, GetItemInfo(SavedItemID)))
						
						if PABdb.profile[abPos]["name"] == SavedItemName and pcall(function() PickupItem(SavedItemID) end) then
							ClearCursor()
							
							PickupItem(SavedItemID)
						end
					end
				end
			end
		end
	
	end -- </LOAD>
end

--- Usage: PAB:pickupFlyouts(1,"save",["placeAction"])
---
--- Usage: PAB:pickupFlyouts(1,"load",["placeAction"])
---
--- @return Save() PABdb.profile[1] = { spellID, altID, icon, version, name, abPos,	}
--- @return Load() PickupSpellBookItem([43},BOOKTYPE_SPELL) (place action if testing)
--- @param actionbarPos	- Actionbars position
--- @param save/load	- Save() to profile[1] / Loads() from profile[1]
--- Description: For easier overview of save and load, converted them into 1 function.
--- 			 Also because the error handling is getting bigger and bigger
function PAB:pickupFlyouts(abPos, command) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if command == "save" then
		local SavedFlyoutaltID  = (select(2, GetActionInfo(abPos)))
		local SavedFlyoutID		= PAB:spellBookIndexbyName(    (select(1, GetFlyoutInfo(SavedFlyoutaltID)))    )
		local SavedFlyoutType 	= (select(1, GetActionInfo(abPos)))
		local SavedFlyoutName 	= (select(1, GetFlyoutInfo(SavedFlyoutaltID)))
		local SavedFlyoutIcon 	= GetActionTexture(abPos)
		
		PickupAction(abPos)
		local SavedFlyoutIconPath = (select(3,GetCursorInfo()))
		PlaceAction(abPos)
		
		PABdb.profile[abPos] = {}
		PABdb.profile[abPos]["type"] 		= SavedFlyoutType
		PABdb.profile[abPos]["ID"] 			= SavedFlyoutID
		PABdb.profile[abPos]["abPos"] 		= abPos
		PABdb.profile[abPos]["name"] 		= SavedFlyoutName
		PABdb.profile[abPos]["version"] 	= ProfilesActionBarsForWoWversion
		PABdb.profile[abPos]["icon"] 		= SavedFlyoutIcon
		PABdb.profile[abPos]["iconPath"] 	= SavedFlyoutIconPath
	end
	
	if command == "load" then
		-- Regular Load()
		if pcall(function() PickupSpellBookItem(PABdb.profile[abPos]["ID"],BOOKTYPE_SPELL) end) then
			ClearCursor()
			PickupSpellBookItem(PABdb.profile[abPos]["ID"],BOOKTYPE_SPELL)
		end
		
		-- Old Load() ID = name
		if not GetCursorInfo() and PABdb.profile[abPos][2] and pcall(function() PickupSpellBookItem(	PAB:spellBookIndexbyName(    PABdb.profile[abPos][4]    ),BOOKTYPE_SPELL) end) then
			ClearCursor()
			
			PickupSpellBookItem(	PAB:spellBookIndexbyName(    PABdb.profile[abPos][4]    ),BOOKTYPE_SPELL)
			
			PAB:NotifyOnce("'"..PABdb:GetCurrentProfile().."': Old profile!", "Warning: This profile contains old records that cannot be loaded in case of version change. |cffee4400Please re-save the profile|r.")
		end
		
		-- Name Load()
		if  not GetCursorInfo() and PABdb.profile[abPos]["name"] then
			ClearCursor()
			
			for FlyoutIndex = 1, GetNumFlyouts() do
				
				if pcall(function() GetFlyoutInfo(FlyoutIndex) end) and PABdb.profile[abPos]["name"] == (select(1, GetFlyoutInfo(FlyoutIndex))) and pcall(function() PickupSpellBookItem(PAB:spellBookIndexbyName(	(select(1, GetFlyoutInfo(FlyoutIndex)))	),BOOKTYPE_SPELL) end) then
					ClearCursor()
					PickupSpellBookItem(PAB:spellBookIndexbyName(	(select(1, GetFlyoutInfo(FlyoutIndex)))	),BOOKTYPE_SPELL) -- Clear AB, Place
				
				end
			end
		end
	end
end

--- Usage: PAB:pickupBattlePet(1,"save",["placeAction"])
---
--- Usage: PAB:pickupBattlePet(1,"load",["placeAction"])
---
--- @return Save() PABdb.profile[1] = { spellID, icon, version, name, abPos,	}
--- @return Load() C_PetJournal.PickupPet("BattlePet-0-00000265DB84") (place action if testing)
--- @param actionbarPos	- Actionbars position
--- @param save/load	- Save() to profile[1] / Loads() from profile[1]
--- Description: For easier overview of save and load, converted them into 1 function.
--- 			 Also because the error handling is getting bigger and bigger
function PAB:pickupBattlePet(abPos, command) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if command == "save" then
		local SavedBattlepetID 		= (select(2,GetActionInfo(abPos)))
		local SavedBattlepetType 	= (select(1, GetActionInfo(abPos)))
		local SavedBattlepetName 	= (select(8, C_PetJournal.GetPetInfoByPetID(SavedBattlepetID)))
		local SavedBattlepetIcon 	= GetActionTexture(abPos)
		
		PABdb.profile[abPos] = {}
		PABdb.profile[abPos]["type"] 		= SavedBattlepetType
		PABdb.profile[abPos]["ID"] 			= SavedBattlepetID
		PABdb.profile[abPos]["abPos"] 		= abPos
		PABdb.profile[abPos]["name"] 		= SavedBattlepetName
		PABdb.profile[abPos]["version"] 	= ProfilesActionBarsForWoWversion
		PABdb.profile[abPos]["icon"] 		= SavedBattlepetIcon
	end
	
	if command == "load" then
		-- Regular load()
		if pcall(function() C_PetJournal.PickupPet(PABdb.profile[abPos]["ID"]) end) then
			ClearCursor()
			C_PetJournal.PickupPet(PABdb.profile[abPos]["ID"])
		end
		
		-- Old load()
		if not GetCursorInfo() and PABdb.profile[abPos][2] and pcall(function() C_PetJournal.PickupPet(PABdb.profile[abPos][2]) end) then
			ClearCursor()
			C_PetJournal.PickupPet(PABdb.profile[abPos][2])
			
			PAB:NotifyOnce("'"..PABdb:GetCurrentProfile().."': Old profile!", "Warning: This profile contains old records that cannot be loaded in case of version change. |cffee4400Please re-save the profile|r.")
		end
		
		-- Name Load()
		if  not GetCursorInfo() and PABdb.profile[abPos]["name"] then
			ClearCursor()
			
			for BattlepetIndex = 1, (select(2, C_PetJournal.GetNumPets())) do
				local SavedBattlepetID 	 = (select(1, C_PetJournal.GetPetInfoByIndex(BattlepetIndex, 	false)))
				local ownedBattlepet 	 = (select(3, C_PetJournal.GetPetInfoByIndex(BattlepetIndex,	false)))
				
				
				if SavedBattlepetID and ownedBattlepet then
					local SavedBattlepetName = (select(8, C_PetJournal.GetPetInfoByPetID(SavedBattlepetID, 	false)))
					
					if PABdb.profile[abPos]["name"] == SavedBattlepetName and pcall(function() C_PetJournal.PickupPet( SavedBattlepetID ) end) then
						ClearCursor()
						C_PetJournal.PickupPet( SavedBattlepetID )
					end
				end
			end
		end
	end
end

--- Usage: PAB:pickupMount(1,"save",["placeAction"])
---
--- Usage: PAB:pickupMount(1,"load",["placeAction"])
---
--- @return Save() PABdb.profile[1] = { MountID, altID, icon, version, name, abPos,	}
--- @return Load() PickupMount(6603) (place action if testing)
--- @param actionbarPos	- Actionbars position
--- @param save/load	- Save() to profile[1] / Loads() from profile[1]
--- Description: For easier overview of save and load, converted them into 1 function.
--- 			 Also because the error handling is getting bigger and bigger
function PAB:pickupMount(abPos, command) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if command == "save" then
		-- companion code
		local SavedMountID 		= (select(2,GetActionInfo(abPos)))
		local SavedMountType 	= (select(1,GetActionInfo(abPos)))
		local SavedMountIcon 	= GetActionTexture(abPos)
		local SavedMountName 	= (select(1, GetSpellInfo(SavedMountID)))
		
		-- summonmount code
		if (SavedMountType == "summonmount" and SavedMountID ~= 268435455) then
			SavedMountName = (select(1, C_MountJournal.GetMountInfoByID(    SavedMountID		)))
			SavedMountID = (select(2, C_MountJournal.GetMountInfoByID(    SavedMountID		))) -- convert from index to spellid
		end
		
		-- Random Favorite Mount code
		if SavedMountID == 268435455 then
			SavedMountName = "Summon Random Favorite Mount"
		end
		
		PABdb.profile[abPos] = {}
		PABdb.profile[abPos]["type"] 		= SavedMountType
		PABdb.profile[abPos]["ID"] 			= SavedMountID
		PABdb.profile[abPos]["abPos"] 		= abPos
		PABdb.profile[abPos]["name"] 		= SavedMountName
		PABdb.profile[abPos]["version"] 	= ProfilesActionBarsForWoWversion
		PABdb.profile[abPos]["icon"] 		= SavedMountIcon
	end
	
	if command == "load" then
		
		-- Regular load()
		if pcall(function() PickupSpell(PABdb.profile[abPos]["ID"]) end) then
			ClearCursor()
			PickupSpell(PABdb.profile[abPos]["ID"])
			
			-- Random mount: Couldn't pickup with PickupSpell(0) from saved, do C_MountJournal.Pickup
			if PABdb.profile[abPos]["ID"] == 268435455 then
				if pcall(function() C_MountJournal.Pickup(0) end) then
					ClearCursor()
					C_MountJournal.Pickup(0)
				end
			end
		end
		
		
		-- Old load()
		if not GetCursorInfo() and pcall(function() C_MountJournal.Pickup(PABdb.profile[abPos][2]) end) then
			ClearCursor()
			local SavedMountID = (select(2,C_MountJournal.GetMountInfoByID(PABdb.profile[abPos][2])))
			C_MountJournal.Pickup(	PAB:GetMountIndex( 	SavedMountID 	))
			
			-- Random mount: Couldn't pickup with PickupSpell(0) from saved, do C_MountJournal.Pickup
			if not GetCursorInfo() and  PABdb.profile[abPos][2] == 268435455 or PABdb.profile[abPos][2] == 0 and pcall(function() C_MountJournal.Pickup(0) end) then
				ClearCursor()
				C_MountJournal.Pickup(0)
			end
			
			PAB:NotifyOnce("'"..PABdb:GetCurrentProfile().."': Old profile!", "Warning: This profile contains old records that cannot be loaded in case of version change. |cffee4400Please re-save the profile|r.")
		end
		
		-- Name Load()
		if not GetCursorInfo() and PABdb.profile[abPos]["name"] then
			ClearCursor()
			
			-- Random mount: Couldn't pickup with PickupSpell(0) from saved, do C_MountJournal.Pickup
			if PABdb.profile[abPos]["name"] == "Summon Random Favorite Mount" and pcall(function() C_MountJournal.Pickup(0) end) then
				ClearCursor()
				C_MountJournal.Pickup(0)
			end
			
			if not GetCursorInfo() then
				for MountIndex = 0, C_MountJournal.GetNumMounts() do
					
					if pcall(function() C_MountJournal.GetMountInfoByID(MountIndex) end) then
						
						local SavedMountName 	= (select(1, C_MountJournal.GetMountInfoByID(MountIndex)))
						local ownedMount 		= (select(11, C_MountJournal.GetMountInfoByID(MountIndex)))	-- Changed for mounts! (select(2, C_MountJournal.GetMountInfoByID(458) pickupablee/not summonable))
						local summonable 		= (select(5, C_MountJournal.GetMountInfoByID(MountIndex)))
						
						if PABdb.profile[abPos]["name"] == SavedMountName and ownedMount and summonable then
							ClearCursor()
							pcall(function() PickupSpell(	(select(2, C_MountJournal.GetMountInfoByID(MountIndex)))	) end)
							
							
							--TODO: Icons don't match, drop
							--local PickedupIcon	=  GetActionTexture(abPos)
							--if PickedupIcon ~= PABdb.profile[abPos]["icon"] and PABdb.profile[abPos]["altID"] ~= 0 then
							--	ClearCursor()
							--end
						end
					end
				end
			end
		end
	
	
	end
end

--- Usage: PAB:pickupEquipmentSet(1,"save",["placeAction"])
---
--- Usage: PAB:pickupEquipmentSet(1,"load",["placeAction"])
---
--- @return Save() PABdb.profile[1] = { spellID (=name), icon, version, name, abPos,	}
--- @return Load() PickupEquipmentSetByName(["test") (place action if testing)
--- @param actionbarPos	- Actionbars position
--- @param save/load	- Save() to profile[1] / Loads() from profile[1]
--- Description: For easier overview of save and load, converted them into 1 function.
--- 			 Also because the error handling is getting bigger and bigger
function PAB:pickupEquipmentSet(abPos, command) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if command == "save" then
		local SavedEquipmentSetID 		= (select(2,GetActionInfo(abPos)))
		local SavedEquipmentSetType 	= (select(1, GetActionInfo(abPos)))
		local SavedEquipmentSetName 	= (select(2,GetActionInfo(abPos)))
		local SavedEquipmentSetIcon 	= GetActionTexture(abPos)
		
		PABdb.profile[abPos] = {}
		PABdb.profile[abPos]["type"] 		= SavedEquipmentSetType
		PABdb.profile[abPos]["ID"] 			= SavedEquipmentSetID
		PABdb.profile[abPos]["abPos"] 		= abPos
		PABdb.profile[abPos]["name"] 		= SavedEquipmentSetName
		PABdb.profile[abPos]["version"] 	= ProfilesActionBarsForWoWversion
		PABdb.profile[abPos]["icon"] 		= SavedEquipmentSetIcon
	
	end
	
	if command == "load" then
		-- Regular load()
		if pcall(function() PickupSpellBookItem(PickupEquipmentSetByName(PABdb.profile[abPos]["ID"])) end) then
			ClearCursor()
			PickupEquipmentSetByName(PABdb.profile[abPos]["ID"])
		end
		
		
		-- Old load()
		if not GetCursorInfo() and pcall(function() PickupSpellBookItem(PickupEquipmentSetByName(PABdb.profile[abPos][2])) end) then
			ClearCursor()
			PickupEquipmentSetByName(PABdb.profile[abPos][2])
			
			PAB:NotifyOnce("'"..PABdb:GetCurrentProfile().."': Old profile!", "Warning: This profile contains old records that cannot be loaded in case of version change. |cffee4400Please re-save the profile|r.")
		end
	end
end

--- Usage: PAB:pickupSpell(1,"save",["placeAction"])
---
--- Usage: PAB:pickupSpell(1,"load",["placeAction"])
---
--- @return Save() PABdb.profile[1] = { spellID, altID, icon, version, name, abPos,	}
--- @return Load() PickupSpell(6603) (place action if testing)
--- @param actionbarPos	- Actionbars position
--- @param save/load	- Save() to profile[1] / Loads() from profile[1]
--- Description: For easier overview of save and load, converted them into 1 function.
--- 			 Also because the error handling is getting bigger and bigger
function PAB:pickupSpell(abPos, command) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if command == "save" then
		
		PickupAction(abPos)
		local SavedSpellID 		= (select(4,GetCursorInfo()))
		PlaceAction(abPos)
		
		local SavedSpellType 	= (select(1, GetActionInfo(abPos)))
		local SavedSpellName 	= (select(1, GetSpellInfo(SavedSpellID)))
		local SavedSpellIcon 	= GetActionTexture(abPos)
		
		PABdb.profile[abPos] = {}
		PABdb.profile[abPos]["type"] 		= SavedSpellType
		PABdb.profile[abPos]["ID"] 			= SavedSpellID
		PABdb.profile[abPos]["abPos"] 		= abPos
		PABdb.profile[abPos]["name"] 		= SavedSpellName
		PABdb.profile[abPos]["version"] 	= ProfilesActionBarsForWoWversion
		PABdb.profile[abPos]["icon"] 		= SavedSpellIcon
	end
	
	if command == "load" then
		-- Regular load()
		if pcall(function() PickupSpell(PABdb.profile[abPos]["ID"]) end) then
			ClearCursor()
			PickupSpell(PABdb.profile[abPos]["ID"])
		end
		
		-- Old load()
		if not GetCursorInfo() and pcall(function() PickupSpell(PABdb.profile[abPos][2]) end) then
			ClearCursor()
			PickupSpell(PABdb.profile[abPos][2])
			
			PAB:NotifyOnce("'"..PABdb:GetCurrentProfile().."': Old profile!", "Warning: This profile contains old records that cannot be loaded in case of version change. |cffee4400Please re-save the profile|r.")
		end
		
		-- Name Load()
		if  not GetCursorInfo() and PABdb.profile[abPos]["name"] then
			ClearCursor()
			
			
			for SpellIndex = 1, MAX_SPELLS do
				local SavedSpellName = (select(1,GetSpellBookItemName(SpellIndex, BOOKTYPE_SPELL)))
				
				if SavedSpellName then
					local SavedSpellID 	 = (select(2, GetSpellBookItemInfo( 	SavedSpellName	)))
					
					if PABdb.profile[abPos]["name"] == SavedSpellName and pcall(function() PickupSpell( SavedSpellID ) end) then
						ClearCursor()
						PickupSpell( SavedSpellID )
					end
				end
			end
		end
		
		
		--- TODO: load by name
	end


end


function PAB:ExistsProfile(name)
	-- Global DB Code
	local DB = {}
	if string.find(name, "General:") then
		DB = PABglobaldb
	else
		DB = PABdb
	end
	
	for i, key in ipairs(DB:GetProfiles()) do
		if key == name then
			return (key)
		end
	end
	
	return nil
end

function PAB:Remove_missing_spells() if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	-- Delete earlier "Missing: *" macros
	for macroID = 1, MAX_ACCOUNT_MACROS do
		
		if string.find((select(1, GetMacroInfo(macroID))) or 1, "Missing: ") then
			DeleteMacro(macroID)
		end
	end
end

--- Usage /pab dump_absPos 1 -> dump whatever is on actionbar 1
---
function PAB:dump_absPos(q, p)
	
	q = tonumber(q) or 1
	
	if not p then
		PAB:Echo("Searching for actionbar nr " .. q .. " in " .. PABdb.char.Active_profile .. "")
	end
	
	for index, keys in pairs(PABdb.profile) do
		if keys["abPos"] == tonumber(q) or keys[3] == tonumber(q) then megadump(keys) end
	end
end



function PAB:ListProfiles(silent, globaldb)
	-- Global DB Code
	local DB = {}
	if globaldb then DB = PABglobaldb
	else DB = PABdb
	end
	
	local PAB_profiles = {}
	
	for i, k in ipairs(DB:GetProfiles()) do
		if not silent then PAB:Echo(' - ' .. k) end
		PAB_profiles[k] = 0
	end
	
	
	return PAB_profiles
end

-- ProfilesHub
function PAB:PreDeleteProfile(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:SendAceMsg("delete", name) -- ProfilesHub
end

-- This is an alias for ProfilesHub
function PAB:Delete(...) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	PAB:DeleteProfil(...)
end

function PAB:DeleteProfil(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	-- Global DB Code
	local DB = {}
	if string.find(name, "General:") then
		DB = PABglobaldb
	else
		DB = PABdb
	end
	
	-- User trying to overwrite restore point!
	if name == "Restore [LOCKED]" and not PABSAVERESTORE then
		PAB:Echo("This profile cannot be deleted! It is locked at the first run and can only be loaded!")
		return nil
	end
	
	if (PAB:ExistsProfile(name) == nil) then
		PAB:Echo(name .. " doesnt exist.")
		PAB:ListProfiles()
		return nil
	else
		if name == DB:GetCurrentProfile() then
			DB:SetProfile('Default')
		end
		DB:DeleteProfile(name)
		PAB:Echo(name .. " deleted.")
		
		return name
	end
end

--- Raw SAVE()
-- This function should NOT BE CALLED DIRECTLY.
-- Should only be called by slave event!
function PAB:SaveABSProfile(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	if name == nil then
		name = PABdb.char.Active_profile
	end
	
	-- Global DB Code
	local DB = {}
	if string.find(name, "General:") then
		DB = PABglobaldb
	else
		DB = PABdb
	end
	
	-- Save a restore point if empty db
	-- PABSAVERESTORE marks "Do not save again" !
	if not PAB:ExistsProfile("Restore [LOCKED]") and not PABSAVERESTORE then
		PABSAVERESTORE = 1
		PAB:SaveABSProfile("Restore [LOCKED]")
		PABSAVERESTORE = nil
	end
	
	-- User trying to overwrite restore point!
	if name == "Restore [LOCKED]" and not PABSAVERESTORE then
		PAB:Echo("This profile is locked at the first run and can only be loaded!")
		return nil
	end
	
	DB:SetProfile(name)
	
	-- Solving /mp save tank X 2
	if PAB:ExistsProfile(name) then DB:ResetProfile(name) end
	
	PAB:Echo("Saving action bar items to profile " .. name .. "!")
	
	for absPos = 1, 120 do
		local SpellType = GetActionInfo(absPos)
		
		if not SpellType then -- NOOP
		elseif (SpellType == "flyout") 		then PAB:pickupFlyouts(absPos,"save")
		elseif (SpellType == "summonpet") 	then PAB:pickupBattlePet(absPos,"save")
		elseif (SpellType == "equipmentset")then PAB:pickupEquipmentSet(absPos,"save")
		elseif (SpellType == "item") 		then PAB:pickupItem(absPos,"save")
		elseif (SpellType == "companion")	then PAB:pickupMount(absPos,"save")
		elseif (SpellType == "summonmount") then PAB:pickupMount(absPos,"save")
		elseif (SpellType == "macro") 		then -- NOOP
		elseif (SpellType == "spell") 		then PAB:pickupSpell(absPos,"save")
		else print("Something went wrong...",SpellType)
		end
	end
	
	return name
end

--- Raw LOAD()
---
--- This function should not be called directly.
-- Should only be called by slave event!
function PAB:LoadABSProfile(name) if select(4, GetBuildInfo()) ~= ProfilesActionBarsForWoWversion and not FORCEPAB then PAB:VersCh() return nil end -- Version block
	
	-- Global DB Code
	local DB = {}
	if string.find(name, "General:") then
		DB = PABglobaldb
	else
		DB = PABdb
	end
	
	-- No such profile saved
	if not PAB:ExistsProfile(name) then PAB:Echo(name .. " profile not found.") return nil end
	
	-- In combat!
	if InCombatLockdown() then PAB:Echo("Cannot change macros in combat!") return end
	
	-- Save a restore point if empty db
	if not PAB:ExistsProfile("Restore [LOCKED]") and not PABSAVERESTORE then
		PABSAVERESTORE = 1
		PAB:SaveABSProfile("Restore [LOCKED]")
		PABSAVERESTORE = nil
	end
	
	
	DB:SetProfile(name)
	PABdb.char.Active_profile = name
	PABdb.char.Version = PABversion
	
	-- Start fresh with nothing on the cursor
	ClearCursor()
	
	-- Create a new "Profile: Macros" profile with missing items
	-- missingActions.SpellNameExtractedFromSaved = spellIDExtractedFromSaved
	-- Macro = print "Profiles: Actionbars - Missing spell: SpellNameExtractedFromSaved / ID spellIDExtractedFromSaved
	local missingActions = {}
	local missingActionsCounter = 0 -- #missingActions wont display correctly..
	
	--Load Action bar profile!
	coreFrameLoopMax = 120
	
	coreFrameLoopIndex = 1
	local coreTimerFrame = CreateFrame("Frame")
	coreTimerFrame.ag = coreTimerFrame:CreateAnimationGroup()
	coreTimerFrame.ag.anim = coreTimerFrame.ag:CreateAnimation()
	coreTimerFrame.ag.anim:SetDuration(PABglobaldb.global.PABloadSpeed)
	coreTimerFrame.ag:SetLooping("REPEAT")
	coreTimerFrame.ag:SetScript("OnLoop",
		function(self)
			if coreFrameLoopIndex == coreFrameLoopMax then
				self:Stop()
				ReleaseFrame(coreTimerFrame)
				PAB:Notify("Loaded profile '" .. name .. "'!","")
				
				--- Create global macros with missing items FIXME
				--- missingActions.SpellNameExtractedFromSaved = spellIDExtractedFromSaved
				--- Macro = print "Profiles: Actionbars - Missing spell: SpellNameExtractedFromSaved / ID spellIDExtractedFromSaved
				if missingActionsCounter > 0 then
					PAB:Echo("DETECTED " .. missingActionsCounter .. " MISSING SPELLS!")
					
					-- Delete earlier "Missing: *" macros
					for macroID = 1, MAX_ACCOUNT_MACROS do
						
						if string.find((select(1, GetMacroInfo(macroID))) or 1, "Missing: ") then
							DeleteMacro(macroID)
						end
					end
					
					-- Do we have space in account macros ?
					if ((MAX_ACCOUNT_MACROS - (select(1, GetNumMacros()))) >= missingActionsCounter) then
						for missingSpellID, missingSpellIDcontainer in pairs(missingActions) do
							
							local spellType = missingSpellIDcontainer[1] or missingSpellIDcontainer["type"] 	or "Unknown"
							local spellName = missingSpellIDcontainer[4] or missingSpellIDcontainer["name"] 	or "Unknown"
							local spellIcon = missingSpellIDcontainer[6] or missingSpellIDcontainer["icon"] 	or 134400
							local spellID 	= missingSpellIDcontainer[2] or missingSpellIDcontainer["ID"] 		or "Unknown ID"
							local actionBar = missingSpellIDcontainer[3] or missingSpellIDcontainer["abPos"] 	or nil
							
							
							
							local createdMacro = CreateMacro("Missing: " .. spellName,
								spellIcon,
								strF("/run PAB:Echo('|%t| cannot be found. Type |%t| ,ID |%t| ,Name |%t| ,Icon |%t| ,Actionbar |%t|')",spellName,spellType,spellID,spellName,spellIcon,actionBar),
								nil,
								nil);
							
							--Clear current spot
							if actionBar then
								ClearCursor()
								PickupAction(actionBar)
								ClearCursor()
								
								PickupMacro(createdMacro)
								PlaceAction(actionBar)
								ClearCursor()
							end
						end
					else
						PAB:Echo("MISSING SPELLS: To be able to show them in GUI please delete " .. missingActionsCounter .. " account macros!")
					end
					
					PAB:Echo("Type '/pab missing_spells' to remove the missing spells macros!")
				end--</Missing macros>
			end
			
			index = coreFrameLoopIndex
			
			ClearCursor() PickupAction(index) ClearCursor() -- Clear
			
			if PABdb.profile[index] then
				if PABdb.profile[index][1] 		and PABdb.profile[index][1] ~= "macro" 		then PickupAction(index) ClearCursor() end -- Do not touch macros!
				if PABdb.profile[index]["type"] and PABdb.profile[index]["type"] ~= "macro" then PickupAction(index) ClearCursor() end -- Do not touch macros!
				
				local PickupSpellType 	= PABdb.profile[index]["type"] or PABdb.profile[index][1]
				
				if 		PickupSpellType == "spell"			then PAB:pickupSpell(index,"load")
				elseif 	PickupSpellType == "flyout" 		then PAB:pickupFlyouts(index,"load")
				elseif 	PickupSpellType == "equipmentset" 	then PAB:pickupEquipmentSet(index,"load")
				elseif 	PickupSpellType == "item" 			then PAB:pickupItem(index,"load")
				elseif 	PickupSpellType == "summonpet" 		then PAB:pickupBattlePet(index,"load")
				elseif 	PickupSpellType == "summonmount" or
						PickupSpellType == "companion" 		then PAB:pickupMount(index,"load")
				end
				
				-- Create a new "Profile: Macros" profile with missing items
				-- missingActions.SpellNameExtractedFromSaved = spellIDExtractedFromSaved
				-- Macro = print "Profiles: Actionbars - Missing spell: SpellNameExtractedFromSaved / ID spellIDExtractedFromSaved
				if not GetCursorInfo() then
					missingActions[index] = PABdb.profile[index]
					missingActionsCounter = missingActionsCounter + 1
				end
				
				PlaceAction(index)
			end -- </If dbrecord not nil>
			
			coreFrameLoopIndex = coreFrameLoopIndex +1
		end)
	coreTimerFrame.ag:Play()
	
	
	PAB:SendMessage("ABS_Profiles_load", name) -- ProfilesHub
	return name
end
